def open_browser(server):
    args = server.cli.parse_known_args()[0]
    local_url = f"http://{args.host}:{server.port}/"
    try:
        import webbrowser
        import asyncio

        loop = asyncio.get_event_loop()
        loop.call_later(0.1, lambda: webbrowser.open(local_url))
        print(
            "And to prevent your browser from opening, "
            "add '--server' to your command line."
        )
    except Exception:
        pass



import json
import sys

from . import logger


def share(obj, attr_name, default_value):
    if obj and hasattr(obj, attr_name):
        return getattr(obj, attr_name)

    return default_value


def _isascii_36(s):
    # For Python < 3.7, we have to use our own isascii() function.
    # This works for both bytes and strings.
    try:
        if isinstance(s, str):
            s.encode("ascii")
        else:
            # Assume bytes
            s.decode("ascii")
    except (UnicodeEncodeError, UnicodeDecodeError):
        return False
    else:
        return True


def _isascii_37(s):
    # For Python >= 3.7, use the built-in function
    return s.isascii()


if sys.version_info >= (3, 7):
    isascii = _isascii_37
else:
    isascii = _isascii_36


def is_dunder(s):
    # Check if this is a double underscore (dunder) name
    return len(s) > 4 and isascii(s) and s[:2] == s[-2:] == "__"


def is_private(s):
    return isascii(s) and s[0] == "_"


def clean_state(state):
    cleaned = {}
    str_values = {}
    for key in state:
        value = clean_value(state[key])
        try:
            str_value = json.dumps(value)
            cleaned[key] = value
            str_values[key] = str_value
        except TypeError:
            logger.error(
                f"Skip state value for '{key}' since its content is not serializable"
            )

    return cleaned, str_values


def clean_value(value):
    if isinstance(value, dict) and "_filter" in value.keys():
        subset = {}
        subset.update(value)
        keys_to_filter = value.get("_filter")
        for key in keys_to_filter:
            subset.pop(key, None)
        return subset

    if isinstance(value, list):
        return list(map(clean_value, value))

    return value


def update_dict(destination, source):
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            container = destination.setdefault(key, {})
            update_dict(container, value)
        else:
            destination[key] = value

    return destination


def reduce_vue_use(state):
    _order = []
    _options = {}
    _reduced = []

    # Merge options
    for item in state.trame__vue_use:
        options = {}
        if isinstance(item, str):
            name = item
        else:
            name, options = item

        _options.setdefault(name, {})
        if name not in _order:
            _order.append(name)

        update_dict(_options[name], options)

    # Generate new content
    for name in _order:
        if len(_options[name]):
            _reduced.append((name, _options[name]))
        else:
            _reduced.append(name)

    # Update state.trame__vue_use with cleaned up version
    state.trame__vue_use = _reduced



import logging
from more_itertools import split_when

logger = logging.getLogger(__name__)

RESERVED_CONTROLLER_NAMES = {
    "on_server_start",
    "on_server_bind",
    "on_server_ready",
    "on_client_connected",
    "on_client_exited",
    "on_server_exited",
    "on_server_reload",
}

JS_DELIMITOR = {
    "'",
    '"',
    ":",
    ";",
    "(",
    ")",
    "{",
    "}",
    " ",
    ".",
    "[",
    "]",
    "`",
    "\n",
    "+",
    "*",
    "/",
    "-",
    "|",
    "&",
}


def js_tokenizer(a, b):
    return a in JS_DELIMITOR or b in JS_DELIMITOR


def vue_template_tokenizer(a, b):
    return a == b and a in {"{", "}"}


def is_name_reserved(name):
    if name.startswith("trame__"):
        return True

    if name in RESERVED_CONTROLLER_NAMES:
        return True

    return False


class Translator:
    """Helper for mapping or namespacing names for state or controller"""

    def __init__(self, prefix=None):
        logger.info("Translator(prefix=%s)", prefix)
        self._prefix = prefix
        self._transl = {}

    def set_prefix(self, prefix):
        self._prefix = prefix

    def add_translation(self, key, translated_key):
        self._transl[key] = translated_key

    def translate_key(self, key):
        # Reserved keys
        if is_name_reserved(key):
            return key

        if key in self._transl:
            return self._transl[key]

        if self._prefix:
            return f"{self._prefix}{key}"

        return key

    def translate_list(self, key_list):
        return [self.translate_key(v) for v in key_list]

    def translate_dict(self, key_dict):
        return {self.translate_key(k): v for k, v in key_dict.items()}

    def translate_js_expression(self, state, expression):
        tokens = []
        for token in split_when(expression, js_tokenizer):
            token = "".join(token)
            logger.info("(prefix=%s) token %s", self._prefix, token)
            if state.has(token):
                _token = self.translate_key(token)
                logger.info("(prefix=%s) translated %s", self._prefix, _token)
                tokens.append(_token)
            else:
                tokens.append(token)

        logger.info(" => %s", "".join(tokens))

        return "".join(tokens)

    def translate_vue_templating(self, state, expression):
        tokens = []
        for token in split_when(expression, vue_template_tokenizer):
            token = "".join(token)
            logger.info(" token %s", token)
            if token.startswith("{"):
                tokens.append(self.translate_js_expression(state, token))
            else:
                tokens.append(token)
        return "".join(tokens)

    def __call__(self, key):
        return self.translate_key(key)



import argparse
import os
import sys


class ArgumentParser(argparse.ArgumentParser):
    """Custom argument parser for Trame

    In this class, `parse_known_args()` has a custom implementation.

    Normally, this would parse sys.argv[1:] as normal, but it will be different
    under the following conditions:

    1. A `--trame-args` argument is provided. If this is provided, then
       only the arguments supplied to this argument will be used for trame.
       For instance: `--trame-args="-p 8081 --server"` will mean the args
       parsed for trame will be `-p 8081 --server`.
    2. A `TRAME_ARGS` environment variable is set. If this is set, then
       only the arguments supplied to this environment variable will be used
       for trame.
       For instance: `TRAME_ARGS="-p 8081 --server"` will mean the args
       parsed for trame will be `-p 8081 --server`.
    3. The `pytest` module has been loaded. If this is the case, then we
       will automatically ignore `sys.argv` because `pytest` will not
       allow for arguments it does not recognize. If `pytest` has been
       loaded, the only way to specify trame arguments is through the
       `TRAME_ARGS` environment variable mentioned in #2.
    """

    def parse_known_args(self, args=None, namespace=None):
        # The trame arguments are usually specified as normal, but they
        # can be alternatively specified via an environment variable or
        # via `--trame-args="..."`.
        args = self._extract_trame_args(args)
        return super().parse_known_args(args, namespace)

    def _extract_trame_args(self, args):
        """Extract the arguments we will parse in trame"""
        if args is not None:
            # If the args is not None, then we are analyzing specific arguments.
            # Just return it the way it is.
            return args

        args = [] if self._skip_default_parsing else sys.argv[1:]

        if any(x.startswith("--trame-args") for x in sys.argv[1:]):
            # Allow argparse to handle all of the different ways this argument may be specified
            tmp_parser = argparse.ArgumentParser()
            tmp_parser.add_argument("--trame-args")
            out, _ = tmp_parser.parse_known_args([v for v in sys.argv[1:] if v != "--"])
            # Add these in
            args += out.trame_args.split()
        elif "TRAME_ARGS" in os.environ:
            # If "--trame-args" wasn't specified, check the environment variable.
            args += os.environ["TRAME_ARGS"].split()

        return args

    @property
    def _skip_default_parsing(self):
        # Skip the default parsing if the trame arguments were set another
        # way, such as the environment variable or the `--trame-args="..."` arg.
        # Also, skip the default parsing if we are in pytest, because pytest
        # will definitely not allow us to have extra arguments anyways.
        return (
            "TRAME_ARGS" in os.environ
            or any(x.startswith("--trame-args") for x in sys.argv[1:])
            or "pytest" in sys.modules
        )



"""Automatically reload function code each time it is called

This module provides a `hot_reload` decorator where the decorated function
is reloaded from source every time the function is called. This works
primarily for global functions and class methods, but it also has
limited support for nested functions as well*.

This module also provides a `reload` function that can be used to reload
a function on demand, i. e. `new_func = reload(old_func)`.

This module is based upon Julian Vossen's reloading library:
https://github.com/julvo/reloading

But it is heavily modified, and includes additional support for things
like class methods. The license for the original library is pasted at
the bottom of this file.

* Nested functions have the following issues:

1. You cannot use the `nonlocal` statement in nested functions. When
the code is reloaded, Python treats it as a global function, and
therefore the `nonlocal` keyword is not allowed.
2. You cannot add capture variables during reloading. Whatever was
captured when the function was first defined is what will continue
to be captured. This is because the outer function is never reloaded,
and the outer function's local scope may potentially be gone at the
time of decorating.
3. If some of the capture variables share the same name as global
variables, but are different, then references to the global variable
within the function or functions it calls may end up referring to
the capture variable instead.
"""

import ast
import functools
import inspect
import site
import sys
import traceback
import types

try:
    from trame_client.ui.core import AbstractLayout
    from trame_client.widgets.core import AbstractElement

    # Skip any methods whose classes inherit from these
    SKIP_CLASSES = [AbstractElement, AbstractLayout]
except ImportError:
    SKIP_CLASSES = []


# Strip any decorators with these names
STRIP_DECORATORS = ["ctrl", "state"]


# We don't actually have logic right now to reload lambdas anyways
SKIP_LAMBDA_FUNCS = True

# If this is True, then skip any functions that are located under any
# site packages directories.
# This essentially means to skip any functions that are not located
# in editable environments.
SKIP_SITE_PACKAGES = True


def hot_reload(func):
    """Decorator to reload the function on every call

    If there are multiple decorators on this function, only the
    decorators after the `@hot_reload` decorator will be reloaded
    """
    if getattr(func, "__is_hot_reload_func", False):
        # It is already a hot_reload function. Just return it.
        # This prevents multiple hot_reload wrappers.
        return func

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # If the user decorated the function manually, make sure we
        # don't perform checks and reload every time.
        new_func = reload(func, perform_checks=False)
        return new_func(*args, **kwargs)

    wrapped.__is_hot_reload_func = True

    return wrapped


def reload(func, perform_checks=True):
    """Attempt to reload the provided function

    This works by locating the file the function was defined in, reloading its
    body of code, and returning the reloaded function.

    If perform_checks is True, then several checks will be performed beforehand
    to determine whether or not the function should be skipped.
    """
    if perform_checks:
        if not isinstance(func, (types.FunctionType, types.MethodType)):
            # We only allow reloading of function and method types
            return func

        if isinstance(func, types.MethodType):
            if isinstance(func.__self__, tuple(SKIP_CLASSES)):
                # We need to skip reloading methods on this class
                return func

        if SKIP_LAMBDA_FUNCS and "<lambda>" in func.__qualname__:
            # Skip lambda functions. It is harder to reload them.
            return func

        if SKIP_SITE_PACKAGES and _func_in_site_packages(func):
            # Skip any packages that were not installed in editable mode
            return func

    while True:
        try:
            return _reload_func(func)
        except Exception:
            _handle_exception(func)


def _reload_func(func):
    func_locals = _find_function_locals(func)
    code = _recompile_function(func)

    # Unfortunately, exec is a little challenging here for non-global
    # functions.
    # Since the source function is compiled by itself, it is treated as
    # a global function, and therefore cannot have closure variables.
    # Because of this, the function will not have read access to the locals
    # we pass in (even though it can still write the function to the locals).
    # To work around this, we copy the locals to the globals (and some globals
    # may be over-written as a result).
    # An alternative work-around would be to add AST code to put a small
    # wrapper function around the function before compiling, such as:
    # def _wrapper():
    #     <func_code>
    # _wrapper()
    # This may get the function to capture the local variables.
    globals_copy = func.__globals__.copy()
    globals_copy.update(func_locals)
    exec(code, globals_copy)

    new_func = globals_copy[func.__name__]

    if isinstance(func, types.MethodType):
        # This is a bound method. Make the new method bound as well.
        new_func = types.MethodType(new_func, func.__self__)

    return new_func


def _find_function_locals(func):
    # First, look at the qualified name.
    # If <locals> is not in the qualified name, then the locals should be
    # the same as the globals. This is much easier for reloading.
    if "<locals>" not in func.__qualname__:
        return func.__globals__

    # If <locals> is in the qualified name, then we may need captured closure
    # variables to run the function correctly. We will use these as the locals.
    # Note that this means that new closure variables cannot be added when
    # reloading the function.
    return inspect.getclosurevars(func).nonlocals


def _recompile_function(func):
    tree = _parse_func_file_until_successful(func)
    if not _isolate_function_def(func.__name__, tree):
        path = inspect.getfile(func)
        raise Exception(f"Failed to find '{func.__qualname__}' in file '{path}'")

    return compile(tree, filename="", mode="exec")


def _parse_func_file_until_successful(func):
    path = inspect.getfile(func)
    while True:
        source = _load_file(path)
        try:
            return ast.parse(source)
        except SyntaxError:
            _handle_exception(func)


def _load_file(path):
    src = ""
    # while loop here since while saving, the file may sometimes be empty.
    while src == "":
        with open(path, "r") as f:
            src = f.read()
    return src + "\n"


def _isolate_function_def(funcname, tree):
    """Strip everything but the function definition from the ast in-place.
    Also strips the hot_reload decorator (including all decorators before it)
    from the function definition"""
    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) and node.name == funcname:
            decorator_names = [_get_decorator_name(dec) for dec in node.decorator_list]
            if "hot_reload" in decorator_names:
                _strip_hot_reload_decorator(node)

            # Remove any decorators that we have indicated to strip
            _strip_decorators(node, STRIP_DECORATORS)

            tree.body = [node]
            return True

    return False


def _handle_exception(func):
    fpath = inspect.getfile(func)
    exc = traceback.format_exc()
    exc = exc.replace('File "<string>"', f'File "{fpath}"')
    sys.stderr.write(exc + "\n")
    print(f"Edit '{func.__qualname__}' in '{fpath}' and press return to continue")
    sys.stdin.readline()


def _get_decorator_name(dec_node):
    if hasattr(dec_node, "id"):
        return dec_node.id
    elif hasattr(dec_node.func, "id"):
        return dec_node.func.id

    return dec_node.func.value.id


def _strip_hot_reload_decorator(func):
    """Remove the 'hot_reload' decorator and all decorators before it"""
    decorator_names = [_get_decorator_name(dec) for dec in func.decorator_list]
    hot_reload_idx = decorator_names.index("hot_reload")
    func.decorator_list = func.decorator_list[hot_reload_idx + 1 :]


def _strip_decorators(func, blacklist):
    """Strip only specific decorators"""
    func.decorator_list = [
        dec for dec in func.decorator_list if _get_decorator_name(dec) not in blacklist
    ]


def _func_in_site_packages(func):
    """Return whether a function is in any site-packages directories"""
    path = inspect.getfile(func)
    return any(path.startswith(x) for x in site.getsitepackages())


JULVO_RELOADING_LICENSE = """MIT License
Copyright (c) 2019 Julian Vossen
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""



import traceback
import aiohttp
import asyncio
import json
from trame_server.utils import asynchronous
from .state import State


def is_attach_key(value):
    return isinstance(value, str) and value.startswith("wslink_bin")


class WsLinkSession:
    CLIENT_ERROR = -32099
    AUTH_ID = "system:c0:0"

    def __init__(self, ws):
        self.loop = asyncio.get_running_loop()
        self.ws = ws
        self.msg_count = 0
        self.bin_id = 1
        self.attachments = {}
        self.subscriptions = {}
        self.client_id = None
        self.in_flight_rpc = {}
        self.in_flight_attachment = []
        self.used_attachment_keys = set()
        self.found_attachment = False

    def _clean_used_attachments(self):
        for key in self.used_attachment_keys:
            self.attachments.pop(key)

        self.used_attachment_keys.clear()

    def _bind_binary(self, msg_result, reset=False):
        if reset:
            self.found_attachment = False

        if msg_result is None:
            return None

        if len(self.attachments):
            if isinstance(msg_result, str) and msg_result in self.attachments:
                self.used_attachment_keys.add(msg_result)
                self.found_attachment = True
                return self.attachments.get(msg_result)
            elif isinstance(msg_result, dict):
                output = {}
                for key, value in msg_result.items():
                    if isinstance(value, (str, dict)):
                        output[key] = self._bind_binary(value, True)
                        if self.found_attachment:
                            output["_filter"] = msg_result.get("_filter", []) + [key]
                    elif key not in output:
                        output[key] = value

                return output

        return msg_result

    async def listen(self):
        async for msg in self.ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                try:
                    payload = json.loads(msg.data)
                except ValueError:
                    print("Malformed TEXT message")
                    continue

                # Notification-only message from the server - should be binary attachment header
                if "id" not in payload:
                    if payload.get("method") == "wslink.binary.attachment":
                        for key in payload.get("args", []):
                            self.attachments[key] = None
                            self.in_flight_attachment.append(key)
                    continue

                msg_id = payload.get("id")
                msg_type, msg_topic, msg_idx = msg_id.split(":")
                future = self.in_flight_rpc.get(msg_id)

                # Error
                if "error" in payload:
                    self.attachments.clear()
                    self.in_flight_attachment.clear()
                    if future:
                        future.set_exception(
                            payload.get("error", "Server error")
                        )  # May need to wrap in Exception?
                    else:
                        print("Server error:", payload.get("error"))

                    self.in_flight_rpc.pop(msg_id)
                    continue

                # Normal processing
                msg_result = payload.get("result")

                # RPC
                if msg_type == "rpc":
                    if future:
                        future.set_result(self._bind_binary(msg_result))
                        self._clean_used_attachments()

                # Publish
                if msg_type == "publish" and msg_topic in self.subscriptions:
                    event = self._bind_binary(msg_result)
                    self._clean_used_attachments()
                    for fn in self.subscriptions[msg_topic]:
                        try:
                            fn(event)
                        except Exception:
                            print("Subscription callback error")
                            traceback.print_exc()

                # System
                if msg_type == "system":
                    if msg_id == WsLinkSession.AUTH_ID:
                        self.client_id = msg_result.get("clientID")
                        future.set_result(self.client_id)
                    else:
                        future.set_result(msg_result)

                # Clean pending future
                if future:
                    self.in_flight_rpc.pop(msg_id)

            elif msg.type == aiohttp.WSMsgType.BINARY:
                if len(self.in_flight_attachment):
                    key = self.in_flight_attachment.pop(0)
                    self.attachments[key] = msg.data
                else:
                    raise ValueError("Attachment missing registration key")
            elif msg.type == aiohttp.WSMsgType.CLOSE:
                print("CLOSE")
            elif msg.type == aiohttp.WSMsgType.CLOSING:
                print("CLOSING")
            elif msg.type == aiohttp.WSMsgType.CLOSED:
                print("CLOSED")
            elif msg.type == aiohttp.WSMsgType.ERROR:
                print("ERROR")
                break

    async def auth(self, **kwargs):
        key = WsLinkSession.AUTH_ID
        resp = self.loop.create_future()
        msg = dict(
            wslink="1.0",
            id=key,
            method="wslink.hello",
            args=[kwargs],
            kwargs={},
        )
        self.in_flight_rpc[key] = resp
        await self.ws.send_str(json.dumps(msg))
        return resp

    async def call(self, method, args=None, kwargs=None):
        self.msg_count += 1
        key = f"rpc:{self.client_id}:{self.msg_count}"
        resp = self.loop.create_future()
        self.in_flight_rpc[key] = resp
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}

        await self.ws.send_str(
            json.dumps(
                dict(
                    wslink="1.0",
                    id=key,
                    method=method,
                    args=args,
                    kwargs=kwargs,
                )
            )
        )

        return resp

    def register_subscription(self, topic, callback):
        if topic not in self.subscriptions:
            self.subscriptions[topic] = [callback]
        else:
            self.subscriptions[topic].append(callback)

    def unregister_subscription(self, topic, callback):
        callbacks = self.subscriptions.get(topic, [])
        if callback in callbacks:
            callbacks.remove(callback)

        if len(callbacks) == 0 and topic in self.subscriptions:
            self.subscriptions.pop(topic)

    def clear_subscriptions(self):
        topics = list(self.subscriptions.keys())
        for topic in topics:
            self.subscriptions.pop(topic)

    async def close(self):
        if self.ws:
            await self.ws.close()


class Client:
    """
    Client implementation for driving a remote trame server with its shared state and trigger method calls in plain python.
    """

    def __init__(self, url=None, config=None, translator=None, hot_reload=False):
        # Network
        self._connected = 0
        self._session = None
        self._url = url
        self._config = {} if config is None else config

        # fake server
        self.hot_reload = hot_reload
        self._change_callbacks = {}

        # trame state
        self._state = State(
            translator, commit_fn=self._push_state, hot_reload=hot_reload
        )

    async def connect(self, url=None, **kwargs):
        if self._connected:
            return
        self._connected = 1

        config = {**self._config, **kwargs}
        if url is None:
            url = self._url

        async with aiohttp.ClientSession() as session:
            async with session.ws_connect(url) as ws:
                self._session = WsLinkSession(ws)
                self._state.ready()
                self._session.register_subscription(
                    "trame.state.topic", self._on_state_update
                )
                self._connected = 2
                task = asynchronous.create_task(self._session.listen())
                await self._session.auth(**config)
                await task

        self._session.clear_subscriptions()
        self._session = None
        self._connected = 0

    async def diconnect(self):
        if self._session:
            await self._session.close()

    # -----------------------------------------------------
    # Fake server for state
    # -----------------------------------------------------

    @property
    def change(self):
        """
        Use as decorator `@server.change(key1, key2, ...)` so the decorated function
        will be called like so `_fn(**state)` when any of the listed key name
        is getting modified from either client or server.

        :param *_args: A list of variable name to monitor
        :type *_args: str
        """
        return self._state.change

    def _push_state(self, state):
        if self._session:
            delta = []
            for key, value in state.items():
                if isinstance(value, dict) and "_filter" in value:
                    skip_keys = set(value.get("_filter"))
                    new_value = {}
                    for k, v in value.items():
                        if k not in skip_keys:
                            new_value[k] = v
                    delta.append(dict(key=key, value=new_value))
                else:
                    delta.append(dict(key=key, value=value))
            asynchronous.create_task(self._session.call("trame.state.update", [delta]))

    def _on_state_update(self, modified_state):
        with self.state:
            self.state.update(modified_state)

    # -----------------------------------------------------

    @property
    def connected(self):
        return self._connected

    @property
    def state(self):
        return self._state

    async def call_trigger(self, name, args=None, kwargs=None):
        if args is None:
            args = []

        if kwargs is None:
            kwargs = {}

        response = await self._session.call("trame.trigger", [name, args, kwargs])
        return await response
